
/* ~~~~~~~~~~~~~~~~~~~~~~~~ Macros ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 

deg, UV

DebugFloat (F, "VariableName")
DebugVector (V, "VariableName")
DebugString (S, "VariableName")
DebugFileDelete ()
DebugFloatPrint (F, "VariableName")
DebugPrintFloat (F, "VariableName")
DebugVectorPrint (V, "VariableName")
DebugPrintVector (V, "VariableName")
DebugVectorPrintXYZ (V)
DebugArrayPrint(Array, iDimFrom, iDimTo, jDimFrom, jDimTo)
PrintDelete ()
PrintReturn ()
PrintFloat (F)
PrintXYZ (V)
PrintVector (V)
PrintString (Str)
PrintDimensions(Object)
GetExt (Str) 

GetFrame (nFrames)

AimAt (vLocation, vDirection, vTarget)
Axes (AxisL)
Axes2 (AxisLength, AxisSize)
Center(Object)
Connect (StartPoint, StartRadius, EndPoint, EndRadius)
ConnectDots  (Dots, nDots)           Note: Dots is a 3-d vector: x, y, z
ConnectDots2 (Dots, nDots)           Note: Dots is a 4-d vector: x, y, z, radius
Distance(V1, V2)
FitTo (vSize1, vSize2, vPos1, vPos2)
FlipXY
FlipXZ
FlipYZ
Flip (Vector, Axis)
Lathe (Array, nPts, Type)
Midpoint(V1, V2)
MinMax (Min, Max, Steps, i) E.g.: MinMax (5, 12, final_frame, frame_number)
MirrorX, MirrorY, MirrorZ
Interpolate (T, T1, T2, P1, P2, E)
CamAngle (Location, Look_at, Width)   - Returns angle in degrees
FillCamera (Location, Look_at, Width)
OrthoCamera (Plane_vector, Width_of_view, Look_at_point) E.g.: OrthoCamera (z, 100, 0)
OrthoCameraOld (Aspect, Plane, ScreenD, ViewW, CenterW, CenterH)
PointLine (P1, P2, Dist)
PointTo (Pos, Target)
Polygon (Array, nPts)
Random (Min, Max)
RandomD (Mean, Deviation, Seed)
RandomS (Min, Max, Seed)
Reorient (Axis1, Axis2)
RotateAxis (Axis, Angle)
RotateCenter (vAngles, vCenter)
RoundedBar (v_SP, v_EP, R)
RoundedBox (v_SP, v_EP ,R)
SmoothRand (Array)
SmoothSeed (Array, Seq)
SwapXY (V), SwapXZ (V), SwapYZ (V), ...
Transition (Start, End, Swerve, Phase)

    ~~~~~~~~~~~~~~~~~~~~~~~~~~ Constants ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

#declare deg = pi/180;  //Angle_in_radians*deg = angle in degress
#declare UV = x+y+z;    //Unit Vector - Vectorizes a float

/*  ~~~~~~~~~~~~~~~~~~~~~~~~ Debug Macros ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#declare OutputFile = "C:/Documents and Settings/Steve/Desktop/Pov Data.xyz"
//#declare OutputFile = "C:\Docume~1\Steve\Desktop\Pov.dat"

#macro DebugFloat (F, VariableName)
  #debug concat(VariableName, " = ", str(F, 0, -1), "\n")
#end
#macro DebugVector (V, VariableName)
  #debug concat(VariableName, " = ", concat("<", str(V.x, 0, -1), ", ", str(V.y, 0, -1), ", ", str(V.z, 0, -1), ">"), "\n")
#end
#macro DebugString (S, VariableName)
  #debug concat(VariableName, " = ", S, "\n")
#end

#macro DebugFileDelete ()
  #fopen File1 OutputFile write
  #fclose File1
#end

#macro DebugFloatPrint (F, VariableName)
  #fopen File1 OutputFile append
  #write (File1, concat(VariableName, " = ", str(F, 0, -1), "\n"))
  #fclose File1
#end
#macro DebugPrintFloat (F, VariableName) DebugFloatPrint (F, VariableName) #end

#macro DebugVectorPrint (V, VariableName)
  #fopen File1 OutputFile append
  #write (File1, concat(VariableName, " = ", concat("<", str(V.x, 0, -1), ", ", str(V.y, 0, -1), ", ", str(V.z, 0, -1), ">"), "\n"))
  #fclose File1
#end
#macro DebugPrintVector (F, VariableName) DebugVectorPrint (F, VariableName) #end

#macro DebugVectorPrintXYZ (V)
  #fopen File1 OutputFile append
  #write (File1, concat(str(V.x, 0, -1), ", ", str(V.y, 0, -1), ", ", str(V.z, 0, -1)), "\n")
  #fclose File1
#end

#macro DebugArrayPrint(Array, iDimFrom, iDimTo, jDimFrom, jDimTo)
  #fopen File1 OutputFile append  
  #local i = iDimFrom; 
  #while (i <= iDimTo)
    #local j = jDimFrom;
    #while (j <= jDimTo)
      //DebugVectorPrint (Array[i][j], concat("[", str(i, 0, 0), "][", str(j, 0, 0), "]"))
      #write (File1, concat(VariableName, " = ", concat("<", str(V.x, 0, -1), ", ", str(V.y, 0, -1), ", ", str(V.z, 0, -1), ">"), "\n"))
      #local j = j + 1;
    #end
    #local i = i + 1;
  #end
  #fclose File1
#end

#macro PrintDelete ()
  #fopen File1 OutputFile write  
  #fclose File1
#end

#macro PrintReturn ()
  #fopen File1 OutputFile append //write  
  #write (File1, "\n")
  #fclose File1
#end

#macro PrintFloat (F)
  #fopen  File1 OutputFile append //write 
  #write (File1, str(F, 0, 7))
  #write (File1, "\n")
  #fclose File1
#end

#macro PrintXYZ (V)
  #fopen  File1 OutputFile append //write 
  #write (File1, vstr(3, V, ", ", 0, 7))
  #write (File1, "\n")
  #fclose File1
#end

#macro PrintVector (V)
  PrintXYZ (V)
#end

#macro PrintString (Str)
  #fopen  File1 OutputFile append
  #write (File1, Str)
  #write (File1, "\n")
  #fclose File1
#end

#macro PrintDimensions(Object)
  PrintReturn()

  PrintString ("Minimum =")
  PrintVector (min_extent(Object))
  PrintString ("Maximum =")
  PrintVector (max_extent(Object))
  
  PrintString ("Width =")
  PrintFloat (max_extent(Object).x-min_extent(Object).x)
  PrintString ("Height =")
  PrintFloat (max_extent(Object).y-min_extent(Object).y)
  PrintString ("Depth =")
  PrintFloat (max_extent(Object).z-min_extent(Object).z)
#end

#macro GetExt (Str)
  substr(Str,strlen(Str)-2,3)
#end

// _________________________ Animation Macros ________________________________

#macro GetFrame (nFrames)
  #local Clock = clock;
  #local Clock_Delta = 1/(nFrames-1);
  #if (Clock = 0)
    #local Frame = 1;
  #else
    #local Frame = (Clock/Clock_Delta)+1;
  #end
  #debug concat("Frame = ", str(Frame, 0, 0), " of ", str(nFrames, 0, 0), "\n")
  #debug concat("Clock = ", str(Clock, 0, -1), ". Going to 1.", "\n")
  
  #if (Clock_Delta != clock_delta & clock_delta != 0 )
    #error "Frame counts do not match."
  #end 
  Frame
  //#debug concat("Clock_Delta = ", str(Clock_Delta, 0, -1), "\n")
  //#debug concat("clock_delta = ", str(clock_delta, 0, -1), "\n")
#end

#macro ClockLoop()  //Cyclical Loop: Clock from 0 to 1 then back to 0 (+ clock_delta)
  #if (mod(final_frame, 2) != 0) #error "Total number of frames must be even" #end // Total frames must be even
  #declare Clock = clock_delta*(frame_number-1);
  #if (Clock > .5) #declare Clock = 1- Clock; #end
  #declare Clock = 2*Clock;
#end
// ____________________________________________________________________________

#macro AimAt (vLocation, vDirection, vTarget) //???
  #local A = vnormalize(vTarget) - vnormalize(vDirection);
  #if (A.x + A.y + A.z != 0)  // If direction and target are not same
    #local vX1 = vnormalize(vDirection);
    #local vX2 = vnormalize(vTarget + (-vLocation));
    #local vY  = vnormalize(vcross(vX2, vX1));
    #local vZ1 = vcross(vY, vX1);
    #local vZ2 = vcross(vY, vX2);
    matrix < vX1.x, vY.x, vZ1.x,
             vX1.y, vY.y, vZ1.y,
             vX1.z, vY.z, vZ1.z,
                 0     0      0>
    matrix < vX2.x, vX2.y, vX2.z,
              vY.x,  vY.y,  vY.z,
             vZ2.x, vZ2.y, vZ2.z,
                 0,     0,     0>
    translate vLocation
  #end 
#end

#macro Axes (AxisL)             //  e.g.: Axes (10)
  #local AxisSize = 0.005;
  #local ArrowL = AxisSize*24;
  #local Axis = union {
  	cylinder {0, (1-ArrowL)*x, AxisSize}
  	cone {x, 0, (1-ArrowL)*x, (tan(radians (15)) * ArrowL)} 
  	no_shadow
  	finish {ambient 0.5}
    scale AxisL
  }
  object {Axis pigment {color rgb x transmit 0.25}}
  object {Axis pigment {color rgb y transmit 0.25} rotate  90*z}
  object {Axis pigment {color rgb z transmit 0.25} rotate -90*y}
  sphere {0, AxisSize pigment {color rgb x transmit 0.25} no_shadow}
#end

#macro Axes2 (AxisLength, AxisSize)
  #ifndef (AxisLength); #declare AxisLength = 10;   #end
  #ifndef (AxisSize):   #declare AxisSize = 0.0025; #end
  
  #declare AxisSize  = AxisSize * AxisLength;
  #declare ArrowSize = AxisSize*12;
  //#declare ArrowSize = AxisSize;
  #declare ArrowAngle  = radians (15);
  #declare ArrowR = tan(ArrowAngle) * ArrowSize;
  #declare Axis = union {
  	cylinder {0, (AxisLength-ArrowSize)*x, AxisSize}
  	cone {AxisLength*x, 0, (AxisLength-ArrowSize)*x, ArrowR} 
  	no_shadow
  	finish {ambient 0.5}
  }
  object {Axis pigment {color rgb x}}
  object {Axis pigment {color rgb y} rotate  90*z}
  object {Axis pigment {color rgb z} rotate -90*y}
  sphere {0, AxisSize pigment {color rgb x} no_shadow}
#end

#macro BevelBox (S, E, BW)
  #local vC=(S+E)/2;
  #local vD=<abs(S.x-E.x),abs(S.y-E.y),abs(S.z-E.z)>/2;
  #local xE=(vD.y+vD.z)/2-BW/sqrt(8);
  #local yE=(vD.x+vD.z)/2-BW/sqrt(8);
  #local zE=(vD.x+vD.y)/2-BW/sqrt(8);
  intersection {
    box { -1,1 matrix <vD.x,0,0, 0,xE,xE, 0,-xE,xE, vC.x,vC.y,vC.z> }
    box { -1,1 matrix <yE,0,-yE, 0,vD.y,0, yE,0,yE, vC.x,vC.y,vC.z> }
    box { -1,1 matrix <zE,zE,0, -zE,zE,0, 0,0,vD.z, vC.x,vC.y,vC.z> }
    bounded_by { box { vC-vD,vC+vD } }
  }
#end

#macro Center(Object)
  min_extent (Object)+max_extent (Object)/2
#end

#macro Connect (P1, R1, P2, R2)
  #local D=vlength(P2-P1);
  #local Rd=R1-R2;
  #local D2=sqrt(D*D-Rd*Rd);
  cone {P1+(P2-P1)/D*Rd*R1/D, R1*D2/D, P2+(P2-P1)/D*Rd*R2/D, R2*D2/D}
#end

#macro ConnectDots (Dots, nDots, r) //Note: Dots[] is an array of 3-d xyz points starting with i = 0
    #local i = 0;
    #while (i < nDots - 1)
      sphere  {<Dots[i].x, Dots[i].y, Dots[i].z>, r}
      Connect (<Dots[i].x, Dots[i].y, Dots[i].z>, r, 
               <Dots[i+1].x, Dots[i+1].y, Dots[i+1].z>, r)  
      #local i = i + 1;
      sphere  {<Dots[i].x, Dots[i].y, Dots[i].z>, r}
    #end
#end

#macro ConnectDots1 (Dots, nDots, r) //Note: Dots[] is an array of 3-d xyz points starting with i = 1
    #local i = 1;
    #while (i < nDots)
      sphere  {<Dots[i].x, Dots[i].y, Dots[i].z>, r}
      Connect (<Dots[i].x, Dots[i].y, Dots[i].z>, r, 
               <Dots[i+1].x, Dots[i+1].y, Dots[i+1].z>, r)  
      #local i = i + 1;
      sphere  {<Dots[i].x, Dots[i].y, Dots[i].z>, r}
    #end
#end

#macro ConnectDots2 (Dots, nDots) //Note: Dots is a 4-d vector: x, y, z, radius. Used to vary line thickness. Index 1
    #local i = 1;           // #local i = 0;
    #while (i < nDots)      // #while (i < nDots - 1)
      sphere  {<Dots[i].x, Dots[i].y, Dots[i].z>, Dots[i].t}
      Connect (<Dots[i].x, Dots[i].y, Dots[i].z>, Dots[i].t, 
               <Dots[i+1].x, Dots[i+1].y, Dots[i+1].z>, Dots[i+1].t)  
      #local i = i + 1;
      sphere  {<Dots[i].x, Dots[i].y, Dots[i].z>, Dots[i].t}
    #end
#end

// Distance between V1 and V2
#macro Distance(V1, V2) vlength(V1 - V2) #end

#macro FitTo (vSize1, vSize2, vPos1, vPos2)
  #local ScaleW = 1; #local ScaleH = 1; #local ScaleD = 1;
  #if (vSize1.x=0 | vSize2.x=0) #else
    #local ScaleW = vSize2.x / vSize1.x;
  #end
  #if (vSize1.y=0 | vSize2.y=0) #else
    #local ScaleH = vSize2.y / vSize1.y;
  #end
  #if (vSize1.z=0 | vSize2.z=0) #else
    #local ScaleD = vSize2.z / vSize1.z;
  #end

  #local dx = vPos2.x - (vPos1.x * ScaleW);
  #local dy = vPos2.y - (vPos1.y * ScaleH);
  #local dz = vPos2.z - (vPos1.z * ScaleD);

  scale <ScaleW, ScaleH, ScaleD>
  translate <dx, dy, dz>
#end

#declare FlipXY = transform { matrix <0,1,0, 1,0,0, 0,0,1, 0,0,0> }
#declare FlipXZ = transform { matrix <0,0,1, 0,1,0, 1,0,0, 0,0,0> }
#declare FlipYZ = transform { matrix <1,0,0, 0,0,1, 0,1,0, 0,0,0> }

#macro Flip (V, Axis)
  #if (Axis.x) <-V.x,  V.y,  V.z> #end
  #if (Axis.y) < V.x, -V.y,  V.z> #end
  #if (Axis.z) < V.x,  V.y, -V.z> #end
#end

#macro Lathe (Array, nPts, Type)
  lathe {
    #if (Type = 1) linear_spline    #end
    #if (Type = 2) quadratic_spline #end
    #if (Type = 3) cubic_spline     #end
    #if (Type = 4) bezier_spline    #end
    nPts,  
    #local i = 1;
    //DebugFloat (i, "i")
    #while (i <= nPts)
      Array[i] 
      #local i = i + 1;
    #end  
    sturm
  }
#end

// Distance between V1 and V2
#macro Midpoint(V1, V2) (V1 + V2) / 2 #end

#macro MinMax (Min, Max, Steps, i)
  #if (initial_frame != 0)
    //#error "MinMax Macro: Initial_Frame must be 0"
  #end
  Min + ((Max - Min) / Steps)*i  
#end

#declare MirrorX = transform {matrix <-1,0,0, 0, 1,0, 0,0, 1, 0,0,0>}
#declare MirrorY = transform {matrix < 1,0,0, 0,-1,0, 0,0, 1, 0,0,0>}
#declare MirrorZ = transform {matrix < 1,0,0, 0, 1,0, 0,0,-1, 0,0,0>}

#macro Interpolate (T, T1, T2, P1, P2, E)
  (P1+(P2-P1)*pow((T-T1)/(T2-T1), E))
#end

#macro CamAngle (Location, Look_at, Width)  //Returns angle in degrees
  #local D = Distance(Location, Look_at);
  atan(Width/D)*180/pi
#end

#macro FillCamera (Location, Look_at, Width)
  camera {
    location  Location
    right     x*image_width/image_height
    look_at   Look_at
    angle CamAngle (Location, Look_at, Width)
  }
#end

#macro OrthoCamera (Plane, ViewW, Look_At)
  camera {
    #local Look_At = Look_At*(x+y+z); //Vectorize if a float
    #local d = 1000; //1E4
    
    orthographic
    #if (abs(Plane.x) = 1)         //Side View
      location  <d*Plane.x, Look_At.y, Look_At.z>
      look_at   <0, Look_At.y, Look_At.z>
    #end
    #if (abs(Plane.y) = 1)         //Top View
      location  <Look_At.x, d*Plane.y, Look_At.z>
      look_at   <Look_At.x, 0, Look_At.z>
    #end
    #if (abs(Plane.z) = 1)         //Front View
      location  <Look_At.x, Look_At.y, -d*Plane.z>
      look_at   <Look_At.x, Look_At.y,  0>
    #end
    right ViewW*x
    up ViewW*y*image_height/image_width 
  }  
#end  

#macro OrthoCameraOld (Aspect, Plane, ScreenD, ViewW, CenterW, CenterH)
  camera {
    #if (ScreenD = 0)
      orthographic
      right    ViewW*x
      up       ViewW*1/Aspect*y 
      #local d = 1E7;
    #else
      right    Aspect*x
      up       y 
      direction ScreenD*z
      #local d = ViewW * (ScreenD/Aspect);
    #end  
    #if (Plane.x = 1)
      location  <d, CenterH, CenterW>
      look_at   <0, CenterH, CenterW>
    #end
    #if (Plane.y = 1)
      location  <CenterW, d, CenterH>
      look_at   <CenterW, 0, CenterH>
    #end
    #if (Plane.z = 1)
      location  <CenterW, CenterH, -d>
      look_at   <CenterW, CenterH,  0>
    #end
  }  
#end  

#macro PointLine (P1, P2, Dist)
  (P1+vnormalize(P2-P1)*Dist)
#end

#macro PointTo (Pos, Target) // Works with surfaces on XY pointing in +Z direction
  #local p = vnormalize(Target + (-Pos)); 
  #if (0+p.x=0 & 0+p.y=0 & 0+p.z=0) 
     #local RotX=0; 
  #else 
     #local RotX=-atan2(p.y,sqrt(pow(p.x,2)+pow(p.z,2)))*180/pi; 
  #end 
  #if (0+p.x=0 & 0+p.z=0) 
     #local RotY=0; 
  #else 
     #local RotY=atan2(p.x,p.z)*180/pi; 
  #end 
  rotate <RotX, RotY, 0> 
  translate Pos 
#end 

#macro Polygon (Array, nPts)
  polygon {nPts,  
    #local i = 1;
    #while (i <= nPts)
      Array[i] 
      #local i = i + 1;
    #end  
  }
#end

#macro Random (Min, Max)
  #ifndef (Rseed) #declare Rseed=seed(1999); #end
  rand(Rseed)*(Max-Min)+Min
#end

#macro RandomD (M, D, Seed)
  (M+(rand(Seed)-.5)*2*D)
#end

#macro RandomS (Min, Max, Seed)
  #local Rseed=seed(Seed);
  rand(Rseed)*(Max-Min)+Min
#end

#macro Reorient(Axis1, Axis2)
  #local vX1=vnormalize(Axis1);
  #local vX2=vnormalize(Axis2);
  #local vY=vnormalize(vcross(vX1,vX2));
  #local vZ1=vnormalize(vcross(vX1,vY));
  #local vZ2=vnormalize(vcross(vX2,vY));
  matrix < vX1.x, vY.x,vZ1.x, vX1.y,vY.y,vZ1.y, vX1.z,vY.z, vZ1.z, 0,0,0 >
  matrix < vX2.x,vX2.y,vX2.z,  vY.x,vY.y, vY.z, vZ2.x,vZ2.y,vZ2.z, 0,0,0 >
#end

#macro RotateAxis (Axis, Angle)
  #local vX=vnormalize(Axis);
  #local vY=vnormalize(vcross(vX,<vX.y,vX.z,-vX.x>));
  #local vZ=vnormalize(vcross(vX,vY));
  matrix < vX.x,vY.x,vZ.x, vX.y,vY.y,vZ.y, vX.z,vY.z,vZ.z, 0,0,0 >
  rotate x*(Angle)
  matrix < vX.x,vX.y,vX.z, vY.x,vY.y,vY.z, vZ.x,vZ.y,vZ.z, 0,0,0 >
#end

#macro RotateCenter (vAngles, vCenter)
  translate -vCenter
  rotate vAngles
  translate vCenter
#end

#macro RoundedBar(v_SP,v_EP,R)
  union {
    #local vSP=(v_SP)*<1,1,1>;
    #local vEP=(v_EP)*<1,1,1>;
    #if (vSP.x > vEP.x)
      #local tSave=vSP.x;
      #local vSP=<vEP.x,vSP.y,vSP.z>;
      #local vEP=<tSave,vEP.y,vEP.z>;
    #end
    #if (vSP.y > vEP.y)
      #local tSave=vSP.y;
      #local vSP=<vSP.x,vEP.y,vSP.z>;
      #local vEP=<vEP.x,sSave,vEP.z>;
    #end
    #if (vSP.z > vEP.z)
      #local tSave=vSP.z;
      #local vSP=<vSP.x,vSP.y,vEP.z>;
      #local vEP=<vEP.x,vEP.y,tSave>;
    #end
  
    prism { vSP.y,vEP.y,9,
      <vEP.x-R,vEP.z>,<vEP.x,vEP.z-R>,<vEP.x,vSP.z+R>,
      <vEP.x-R,vSP.z>,<vSP.x+R,vSP.z>,<vSP.x,vSP.z+R>,
      <vSP.x,vEP.z-R>,<vSP.x+R,vEP.z>,<vEP.x-R,vEP.z> }
    cylinder { <vSP.x+R,vSP.y,vSP.z+R>,<vSP.x+R,vEP.y,vSP.z+R>,R }
    cylinder { <vEP.x-R,vSP.y,vSP.z+R>,<vEP.x-R,vEP.y,vSP.z+R>,R }
    cylinder { <vSP.x+R,vSP.y,vEP.z-R>,<vSP.x+R,vEP.y,vEP.z-R>,R }
    cylinder { <vEP.x-R,vSP.y,vEP.z-R>,<vEP.x-R,vEP.y,vEP.z-R>,R }
  }
#end

#macro RoundedBox(v_SP,v_EP,R)
  union {
    #local vSP=(v_SP)*<1,1,1>;
    #local vEP=(v_EP)*<1,1,1>;
    #if (vSP.x > vEP.x)
      #local tSave=vSP.x;
      #local vSP=<vEP.x,vSP.y,vSP.z>;
      #local vEP=<tSave,vEP.y,vEP.z>;
    #end
    #if (vSP.y > vEP.y)
      #local tSave=vSP.y;
      #local vSP=<vSP.x,vEP.y,vSP.z>;
      #local vEP=<vEP.x,sSave,vEP.z>;
    #end
    #if (vSP.z > vEP.z)
      #local tSave=vSP.z;
      #local vSP=<vSP.x,vSP.y,vEP.z>;
      #local vEP=<vEP.x,vEP.y,tSave>;
    #end
  
    box {<vSP.x,vSP.y+R,vSP.z+R>,<vEP.x,vEP.y-R,vEP.z-R>}
    box {<vSP.x+R,vSP.y,vSP.z+R>,<vEP.x-R,vEP.y,vEP.z-R>}
    box {<vSP.x+R,vSP.y+R,vSP.z>,<vEP.x-R,vEP.y-R,vEP.z>}
  
    #local vSP=vSP+R;
    #local vEP=vEP-R;
  
    sphere { <vSP.x,vSP.y,vSP.z>,R}
    sphere { <vEP.x,vSP.y,vSP.z>,R}
    sphere { <vSP.x,vEP.y,vSP.z>,R}
    sphere { <vEP.x,vEP.y,vSP.z>,R}
    sphere { <vSP.x,vSP.y,vEP.z>,R}
    sphere { <vEP.x,vSP.y,vEP.z>,R}
    sphere { <vSP.x,vEP.y,vEP.z>,R}
    sphere { <vEP.x,vEP.y,vEP.z>,R}
  
    cylinder { <vSP.x,vSP.y,vSP.z>,<vEP.x,vSP.y,vSP.z>,R }
    cylinder { <vSP.x,vEP.y,vSP.z>,<vEP.x,vEP.y,vSP.z>,R }
    cylinder { <vSP.x,vSP.y,vEP.z>,<vEP.x,vSP.y,vEP.z>,R }
    cylinder { <vSP.x,vEP.y,vEP.z>,<vEP.x,vEP.y,vEP.z>,R }
    cylinder { <vSP.x,vSP.y,vSP.z>,<vSP.x,vEP.y,vSP.z>,R }
    cylinder { <vEP.x,vSP.y,vSP.z>,<vEP.x,vEP.y,vSP.z>,R }
    cylinder { <vSP.x,vSP.y,vEP.z>,<vSP.x,vEP.y,vEP.z>,R }
    cylinder { <vEP.x,vSP.y,vEP.z>,<vEP.x,vEP.y,vEP.z>,R }
    cylinder { <vSP.x,vSP.y,vSP.z>,<vSP.x,vSP.y,vEP.z>,R }
    cylinder { <vEP.x,vSP.y,vSP.z>,<vEP.x,vSP.y,vEP.z>,R }
    cylinder { <vSP.x,vEP.y,vSP.z>,<vSP.x,vEP.y,vEP.z>,R }
    cylinder { <vEP.x,vEP.y,vSP.z>,<vEP.x,vEP.y,vEP.z>,R }
  }
#end

#macro BiRoundedBox(Start, End, MajorRadius, MinorRadius)
  #if (MajorRadius=MinorRadius)
    RoundedBox(Start,End,MajorRadius)
  #end
  #if (MajorRadius>MinorRadius)
    #if (MinorRadius=0)
      RoundedBar(Start,End,MajorRadius)
    #else
      #local pS=Start;
      #local pE=End;
      #if (pS.x > pE.x)
        #local tSave=pS.x;
        #local pS=<pE.x,pS.y,pS.z>;
        #local pE=<tSave,pE.y,pE.z>;
      #end
      #if (pS.y > pE.y)
        #local tSave=pS.y;
        #local pS=<pS.x,pE.y,pS.z>;
        #local pE=<pE.x,tSave,pE.z>;
      #end
      #if (pS.z > pE.z)
        #local tSave=pS.z;
        #local pS=<pS.x,pS.y,pE.z>;
        #local pE=<pE.x,pE.y,tSave>;
      #end
      #local rY=abs(MajorRadius);
      #local rXZ=abs(MinorRadius);

      RoundedBar(pS+<0,rXZ,0>,   pE-<0,rXZ,0>,  rY)
      RoundedBar(pS+<rXZ,0,rXZ>,pE-<rXZ,0,rXZ>,rY-rXZ)

      torus { rY-rXZ,rXZ translate <pS.x+rY,pS.y+rXZ,pS.z+rY> }
      torus { rY-rXZ,rXZ translate <pE.x-rY,pS.y+rXZ,pS.z+rY> }
      torus { rY-rXZ,rXZ translate <pS.x+rY,pE.y-rXZ,pS.z+rY> }
      torus { rY-rXZ,rXZ translate <pE.x-rY,pE.y-rXZ,pS.z+rY> }
      torus { rY-rXZ,rXZ translate <pS.x+rY,pS.y+rXZ,pE.z-rY> }
      torus { rY-rXZ,rXZ translate <pE.x-rY,pS.y+rXZ,pE.z-rY> }
      torus { rY-rXZ,rXZ translate <pS.x+rY,pE.y-rXZ,pE.z-rY> }
      torus { rY-rXZ,rXZ translate <pE.x-rY,pE.y-rXZ,pE.z-rY> }
      
      cylinder { <pS.x+rY,pS.y+rXZ,pS.z+rXZ>,<pE.x-rY,pS.y+rXZ,pS.z+rXZ>,rXZ }
      cylinder { <pS.x+rY,pE.y-rXZ,pS.z+rXZ>,<pE.x-rY,pE.y-rXZ,pS.z+rXZ>,rXZ }
      cylinder { <pS.x+rY,pS.y+rXZ,pE.z-rXZ>,<pE.x-rY,pS.y+rXZ,pE.z-rXZ>,rXZ }
      cylinder { <pS.x+rY,pE.y-rXZ,pE.z-rXZ>,<pE.x-rY,pE.y-rXZ,pE.z-rXZ>,rXZ }
  
      cylinder { <pS.x+rXZ,pS.y+rXZ,pS.z+rY>,<pS.x+rXZ,pS.y+rXZ,pE.z-rY>,rXZ }
      cylinder { <pE.x-rXZ,pS.y+rXZ,pS.z+rY>,<pE.x-rXZ,pS.y+rXZ,pE.z-rY>,rXZ }
      cylinder { <pS.x+rXZ,pE.y-rXZ,pS.z+rY>,<pS.x+rXZ,pE.y-rXZ,pE.z-rY>,rXZ }
      cylinder { <pE.x-rXZ,pE.y-rXZ,pS.z+rY>,<pE.x-rXZ,pE.y-rXZ,pE.z-rY>,rXZ }
    #end
  #end
#end

#macro SmoothRand(Array)
  #local t_T=clock-floor(clock);
  #local t_R0=rand(Array[0]);
  #local t_R1=rand(Array[1]);
  #local t_R2=rand(Array[2]);
  #local t_R3=rand(Array[3]);
  #local t_C0=t_R1;
  #local t_C1=(t_R2-t_R0)*.5;
  #local t_C2=t_R0 - 2.5*t_R1 +2*t_R2-.5*t_R3;
  #local t_C3=-.5*t_R0 +1.5*t_R1 -1.5*t_R2 +.5*t_R3;
  (((((t_C3*t_T+t_C2)*t_T+t_C1)*t_T+t_C0)*.8)+.1)
#end

#macro SmoothSeed(Array, Seq)
  #declare Array[0]= seed(floor(clock-1)+Seq);
  #declare Array[1]= seed(floor(clock)+Seq);
  #declare Array[2]= seed(floor(clock+1)+Seq);
  #declare Array[3]= seed(floor(clock+2)+Seq);
#end

#macro SwapXY(V) <V.y, V.x, V.z> #end
#macro SwapYX(V) SwapXY(V) #end
#macro SwapXZ(V) <V.z, V.y, V.x> #end
#macro SwapZX(V) SwapXZ(V) #end
#macro SwapYZ(V) <V.x, V.z, V.y> #end
#macro SwapZY(V) SwapYZ(V) #end

#macro Transition(Start, End, Swerve, Phase)
  #local t_P=(Phase)*pi;
  #local t_TL=(1-cos(t_P))/2;
  #local t_TS=(1-cos(t_P*2))/2;
  ( (Start)*(1-t_TL) + (End)*t_TL + (Swerve)*t_TS )
#end

